#!/usr/bin/env python3
##############################################################################
#
# Copyright (c) 2022 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import argparse
import collections
import configparser
import os
import pathlib
import shutil
import sys

import tomlkit

from .shared.call import call
from .shared.call import wait_for_accept
from .shared.git import get_branch_name
from .shared.git import git_branch
from .shared.packages import FUTURE_PYTHON_VERSION
from .shared.packages import OLDEST_PYTHON_VERSION
from .shared.packages import supported_python_versions
from .shared.path import change_dir


def get_tox_ini_python_versions(path) -> set:
    config = configparser.ConfigParser()
    config.read(path)
    envs = config['tox']['envlist'].split()
    raw_versions = {
        env.replace('py3', '3.') for env in envs
        if env.startswith('py') and env != 'pypy3'
    }
    # c-code template uses `py313,py313-pure`, get rid of the pure version:
    versions = {v.partition(',')[0] for v in raw_versions}
    # some packages use `py38-watch` or similar, get rid of the suffix:
    return {v.partition('-')[0] for v in versions}


def main():
    parser = argparse.ArgumentParser(
        description='Update Python versions of a package to currently '
        'supported ones.')
    parser.add_argument('path',
                        type=pathlib.Path,
                        help='path to the repository to be configured')
    parser.add_argument(
        '--branch',
        dest='branch_name',
        default=None,
        help='Define a git branch name to be used for the changes. If not'
        ' given it is constructed automatically and includes the configuration'
        ' type')
    parser.add_argument(
        '--no-commit',
        dest='commit',
        action='store_false',
        default=True,
        help='Don\'t "git commit" changes made by this script.')
    parser.add_argument(
        '--with-future-python',
        dest='with_future_python',
        action='store_true',
        default=False,
        help='Also enable testing the future Python version.')
    parser.add_argument(
        '--interactive',
        dest='interactive',
        action='store_true',
        default=False,
        help='Run interactively: Scripts will prompt for input. Implies '
        '--no-commit, changes will not be committed and pushed automatically.')
    parser.add_argument(
        '--no-tests',
        dest='run_tests',
        action='store_false',
        default=True,
        help='Skip running unit tests.')
    parser.add_argument(
        '--template-overrides',
        dest='template_override_path',
        default=None,
        help='Filesystem path to a folder with subfolders for configuration '
        'types. Used to override built-in configuration templates.')

    args = parser.parse_args()
    path = args.path.absolute()

    if not (path / '.git').exists():
        raise ValueError(
            '`path` does not point to a git clone of a repository!')
    if not (path / '.meta.toml').exists():
        raise ValueError('The repository `path` points to has no .meta.toml!')

    with change_dir(path) as cwd_str:
        cwd = pathlib.Path(cwd_str)
        bin_dir = cwd / 'bin'
        with open('.meta.toml', 'rb') as meta_f:
            meta_toml = collections.defaultdict(dict, **tomlkit.load(meta_f))
        config_type = meta_toml['meta']['template']
        oldest_python_version = meta_toml['python'].get('oldest-python',
                                                        OLDEST_PYTHON_VERSION)
        branch_name = get_branch_name(args.branch_name, config_type)
        updating = git_branch(branch_name)

        current_python_versions = get_tox_ini_python_versions('tox.ini')
        no_longer_supported = (
            current_python_versions -
            set(supported_python_versions(oldest_python_version))
        )
        not_yet_supported = (
            set(supported_python_versions(oldest_python_version)) -
            current_python_versions
        )

        non_interactive_params = []
        python_versions_args = []
        if not args.interactive and args.commit:
            non_interactive_params = ['--no-input']
        else:
            args.commit = False

        if no_longer_supported or not_yet_supported:
            call(bin_dir / 'bumpversion', '--feature', *non_interactive_params)
        else:
            print('No changes required.')
            sys.exit(0)

        if no_longer_supported:
            version_spec = ', '.join(sorted(no_longer_supported))
            call(bin_dir / 'addchangelogentry',
                 f'Drop support for Python {version_spec}.',
                 *non_interactive_params)
            python_versions_args.append(
                '--drop=' + ','.join(no_longer_supported))

        if not_yet_supported:
            version_spec = ', '.join(sorted(not_yet_supported))
            call(
                bin_dir / 'addchangelogentry',
                f'Add support for Python {version_spec}.',
                *non_interactive_params)
            python_versions_args.append(
                '--add=' +
                ','.join(supported_python_versions(oldest_python_version))
            )
        if args.with_future_python:
            call(
                bin_dir / 'addchangelogentry',
                f'Add preliminary support for Python {FUTURE_PYTHON_VERSION}.',
                *non_interactive_params)

        if no_longer_supported or not_yet_supported:
            call(bin_dir / 'check-python-versions', '--only=setup.py',
                 *python_versions_args)
            print('Look through .meta.toml to see if it needs changes.')
            call(os.environ['EDITOR'], '.meta.toml')

            config_package_args = [
                bin_dir / 'config-package',
                path,
                f'--branch={branch_name}',
                '--no-push',
            ]
            if args.with_future_python:
                config_package_args.append('--with-future-python')
            if not args.commit:
                config_package_args.append('--no-commit')
            if not args.run_tests:
                config_package_args.append('--no-tests')
            if args.template_override_path:
                config_package_args.append(
                    f'--template-overrides={args.template_override_path}')
            call(*config_package_args, cwd=cwd_str)
            src = path.resolve() / 'src'
            py_ver_plus = f'--py{oldest_python_version.replace(".", "")}-plus'
            call('find', src, '-name', '*.py', '-exec', bin_dir / 'pyupgrade',
                 '--py3-plus', py_ver_plus, '{}', ';')
            call(bin_dir / 'pyupgrade',
                 '--py3-plus',
                 py_ver_plus,
                 'setup.py',
                 allowed_return_codes=(0, 1))

            excludes = (
                '--exclude-dir',
                '__pycache__',
                '--exclude-dir',
                '*.egg-info',
                '--exclude',
                '*.pyc',
                '--exclude',
                '*.so')
            print('Replace any remaining code that might'
                  ' support legacy Python:')
            call(
                'egrep',
                '-rn',
                f'{"|".join(no_longer_supported)}|sys.version|PY3|Py3|Python 3'
                '|__unicode__|ImportError',
                src,
                *excludes,
                allowed_return_codes=(
                    0,
                    1))
            wait_for_accept()

            if args.run_tests:
                tox_path = shutil.which('tox') or (cwd / 'bin' / 'tox')
                call(tox_path, '-p', 'auto')

            if args.commit:
                print('Adding, committing and pushing all changes ...')
                call('git', 'add', '.')
                call('git', 'commit', '-m', 'Update Python version support.')
                call('git', 'push', '--set-upstream', 'origin', branch_name)
                if updating:
                    print('Updated the previously created PR.')
                else:
                    print(
                        'Are you logged in via `gh auth login` to'
                        ' create a PR? (y/N)?', end=' ')
                    if input().lower() == 'y':
                        call('gh', 'pr', 'create', '--fill', '--title',
                             'Update Python version support.')
                    else:
                        print('If everything went fine up to here:')
                        print('Create a PR, using the URL shown above.')
            else:
                print('Applied all changes. Please check and commit manually.')



#!/usr/bin/env python3
##############################################################################
#
# Copyright (c) 2021 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import argparse
import pathlib

from .shared.call import call
from .shared.packages import ALL_REPOS
from .shared.packages import ORG


base_url = f'https://github.com/{ORG}'
BASE_PATH = pathlib.Path(__file__).parent


def run_workflow(base_url, org, repo):
    """Manually start the tests.yml workflow of a repository."""
    result = call('gh', 'workflow', 'run', 'tests.yml', '-R', f'{org}/{repo}')
    if result.returncode != 0:
        print('To enable manually starting workflows clone the repository'
              ' and run meta/config/config-package.py on it.')
        print('Command to clone:')
        print(f'git clone {base_url}/{repo}.git')
        return False
    return True


def main():
    parser = argparse.ArgumentParser(
        description='Re-enable GitHub Actions for all repos in a packages.txt'
                    ' files.')
    parser.add_argument(
        '--force-run',
        help='Run workflow even it is already enabled.',
        action='store_true')

    args = parser.parse_args()

    for repo in ALL_REPOS:
        print(repo)
        wfs = call(
            'gh', 'workflow', 'list', '--all', '-R', f'{ORG}/{repo}',
            capture_output=True).stdout
        try:
            test_line = [
                x for x in wfs.splitlines() if x.startswith('test')][0]
        except IndexError:
            print('    ❌ no tests.yml workflow')
            continue
        if 'disabled_inactivity' not in test_line:
            print('    ☑️  already enabled')
            if args.force_run:
                run_workflow(base_url, ORG, repo)
            continue
        test_id = test_line.split()[-1]
        call('gh', 'workflow', 'enable', test_id, '-R', f'{ORG}/{repo}')
        if run_workflow(base_url, ORG, repo):
            print('    ✅ enabled')



##############################################################################
#
# Copyright (c) 2025 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import tempfile
import textwrap
import unittest

from zope.meta.update_python_support import get_tox_ini_python_versions


class TestUpdatePythonSupport(unittest.TestCase):

    def test_update_python_support__get_tox_ini_python_versions__1(self):
        """It filters out c-code pure versions and custom versions."""

        with tempfile.NamedTemporaryFile(mode='w') as tox_ini:
            tox_ini.write(textwrap.dedent("""
                [tox]
                minversion = 4.0
                envlist =
                    lint
                    py39,py39-pure
                    py310,py310-pure
                    py311,py311-pure
                    pypy3
                    docs
                    coverage
                    py39-watch, py311-watch
                """))
            tox_ini.flush()
            versions = get_tox_ini_python_versions(tox_ini.name)
            self.assertEqual({'3.9', '3.10', '3.11'}, versions)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import datetime


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
year = datetime.datetime.now().year

project = 'zope.meta'
copyright = f'2020-{year}, Zope Foundation and contributors'
author = 'Zope Foundation and contributors'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
]

# templates_path = ['_templates']
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'furo'
# html_static_path = ['_static']

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'testrunner': ('https://zopetestrunner.readthedocs.io/en/stable/', None),
    'coverage': ('https://coverage.readthedocs.io/en/latest', None),
}



##############################################################################
#
# Copyright (c) 2024 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest


class ConfigPackageTests(unittest.TestCase):

    def test_prepend_space(self):
        from zope.meta.config_package import prepend_space

        self.assertIsNone(prepend_space(None))
        self.assertEqual('', prepend_space(''))
        self.assertEqual(' foobar', prepend_space('foobar'))



#!/usr/bin/env python3
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import argparse
import collections
import pathlib
import re
import shutil
from functools import cached_property

import jinja2
import tomlkit
from packaging.version import InvalidVersion
from packaging.version import parse as parse_version

from .set_branch_protection_rules import set_branch_protection
from .shared.call import abort
from .shared.call import call
from .shared.git import get_branch_name
from .shared.git import get_commit_id
from .shared.git import git_branch
from .shared.packages import FUTURE_PYTHON_VERSION
from .shared.packages import MANYLINUX_AARCH64
from .shared.packages import MANYLINUX_I686
from .shared.packages import MANYLINUX_PYTHON_VERSION
from .shared.packages import MANYLINUX_X86_64
from .shared.packages import NEWEST_PYTHON_VERSION
from .shared.packages import OLDEST_PYTHON_VERSION
from .shared.packages import PYPY_VERSION
from .shared.packages import SETUPTOOLS_VERSION_SPEC
from .shared.packages import get_pyproject_toml
from .shared.packages import parse_additional_config
from .shared.packages import supported_python_versions
from .shared.path import change_dir


FUTURE_PYTHON_SHORTVERSION = FUTURE_PYTHON_VERSION.replace('.', '')
NEWEST_PYTHON_SHORTVERSION = NEWEST_PYTHON_VERSION.replace('.', '')
META_HINT = """\
# Generated from:
# https://github.com/zopefoundation/meta/tree/master/config/{config_type}"""
META_HINT_MARKDOWN = """\
<!--
Generated from:
https://github.com/zopefoundation/meta/tree/master/config/{config_type}
-->"""
DEFAULT = object()
SETUP_PY_REPLACEMENTS = {
    'ZPL 2.1': 'ZPL-2.1',
    'zope-dev@zope.org': 'zope-dev@zope.dev',
    'Zope Corporation': 'Zope Foundation',
    'Framework :: Zope2': 'Framework :: Zope :: 2',
    'Framework :: Zope3': 'Framework :: Zope :: 3',
}


def handle_command_line_arguments():
    """Parse command line options"""
    parser = argparse.ArgumentParser(
        description='Use configuration for a package.')
    parser.add_argument(
        'path', type=pathlib.Path,
        help='path to the repository to be configured')
    parser.add_argument(
        '--commit-msg',
        dest='commit_msg',
        metavar='MSG',
        help='Use MSG as commit message instead of an artificial one.')
    parser.add_argument(
        '--no-commit',
        dest='commit',
        action='store_false',
        default=True,
        help='Prevent automatic committing of changes. Implies --no-push.')
    parser.add_argument(
        '--no-push',
        dest='push',
        action='store_false',
        default=True,
        help='Prevent direct push.')
    parser.add_argument(
        '--no-tests',
        dest='run_tests',
        action='store_false',
        default=True,
        help='Skip running unit tests.')
    parser.add_argument(
        '--with-macos',
        dest='with_macos',
        action='store_true',
        default=False,
        help='Activate running tests on macOS on GitHub Actions, too, if not'
        ' already configured in .meta.toml.')
    parser.add_argument(
        '--with-windows',
        dest='with_windows',
        action='store_true',
        default=False,
        help='Activate running tests on Windows on GitHub Actions, too, if not'
        ' already configured in .meta.toml.')
    parser.add_argument(
        '--with-pypy',
        dest='with_pypy',
        action='store_true',
        default=False,
        help='Activate PyPy support if not already configured in .meta.toml.')
    parser.add_argument(
        '--with-future-python',
        dest='with_future_python',
        action='store_true',
        default=False,
        help='Activate support for a future non-final Python version if not'
        ' already configured in .meta.toml.')
    parser.add_argument(
        '--oldest-python',
        dest='oldest_python',
        help='Oldest supported Python version. Defaults to:'
             f' {OLDEST_PYTHON_VERSION}.')
    parser.add_argument(
        '--with-docs',
        # people (me) use --with-sphinx and accidentally
        # get --with-sphinx-doctests
        # so let's make --with-sphinx an alias for --with-docs
        '--with-sphinx',
        dest='with_docs',
        action='store_true',
        default=False,
        help='Activate building docs if not already configured in .meta.toml.')
    parser.add_argument(
        '--with-sphinx-doctests',
        dest='with_sphinx_doctests',
        action='store_true',
        default=False,
        help='Activate running doctests with sphinx '
        'if not already configured in  .meta.toml.')
    parser.add_argument(
        '-t', '--type',
        choices=[
            'buildout-recipe',
            'c-code',
            'pure-python',
            'zope-product',
            'toolkit',
        ],
        default=None,
        dest='type',
        help='type of the configuration to be used, see README.rst. '
        'Only required when running on a repository for the first time.')
    parser.add_argument(
        '--branch',
        dest='branch_name',
        default=None,
        help='Define a git branch name to be used for the changes. '
        'If not given it is constructed automatically and includes '
        'the configuration type')
    parser.add_argument(
        '--template-overrides',
        dest='template_override_path',
        default=None,
        help='Filesystem path to a folder with subfolders for configuration '
        'types. Used to override built-in configuration templates.')

    args = parser.parse_args()
    return args


def prepend_space(text):
    """Prepend `text` with a space if not empty.

    This prevents trailing whitespace for empty values.
    """
    if text:
        text = f' {text}'
    return text


class PackageConfiguration:
    add_manylinux = False

    def __init__(self, args):
        self.args = args
        self.path = args.path.absolute()
        self.meta_cfg = {}

        if not (self.path / '.git').exists():
            raise ValueError(
                f'{self.path!r} does not point '
                'to a git clone of a repository!')

        self.meta_cfg = self._read_meta_configuration()
        self.meta_cfg['meta']['template'] = self.config_type
        self.meta_cfg['meta']['commit-id'] = get_commit_id()

    def _read_meta_configuration(self):
        """Read and update meta configuration"""
        meta_toml_path = self.path / '.meta.toml'
        if meta_toml_path.exists():
            with open(meta_toml_path, 'rb') as meta_f:
                meta_cfg = tomlkit.load(meta_f)
            meta_cfg = collections.defaultdict(dict, **meta_cfg)
        else:
            meta_cfg = collections.defaultdict(dict)
            if self.args.with_docs is None:
                self.args.with_docs = (self.path / "docs" / "conf.py").exists()
                print(f"Autodetecting --with-docs: {self.args.with_docs}")
        return meta_cfg

    def template_exists(self, filename):
        """Check if a given template exists"""
        for parent_folder in self.template_folders:
            if (parent_folder / filename).exists():
                return True
        return False

    @cached_property
    def config_type(self):
        value = self.meta_cfg['meta'].get('template') or self.args.type
        if value is None:
            raise ValueError(
                'Configuration type not set. '
                'Please use `--type` to select it.')
        return value

    @cached_property
    def oldest_python(self):
        value = (self.args.oldest_python or
                 self.meta_cfg['python'].get('oldest-python') or
                 OLDEST_PYTHON_VERSION)
        try:
            version = parse_version(value)
        except InvalidVersion:
            raise ValueError(f'Invalid value {value} for oldest Python.')

        if version > parse_version(NEWEST_PYTHON_VERSION):
            raise ValueError('Oldest Python version cannot be higher than'
                             ' newest supported Python')

        return value

    @cached_property
    def config_type_path(self):
        return pathlib.Path(__file__).parent / self.config_type

    @cached_property
    def default_path(self):
        return pathlib.Path(__file__).parent / 'default'

    @cached_property
    def override_paths(self):
        if self.args.template_override_path:
            override_path = pathlib.Path(self.args.template_override_path)
            return [override_path / self.config_type,
                    override_path / 'default']
        return []

    @cached_property
    def template_folders(self):
        return self.override_paths + [self.config_type_path, self.default_path]

    @cached_property
    def jinja_env(self):
        return jinja2.Environment(
            loader=jinja2.FileSystemLoader(self.template_folders),
            variable_start_string='%(',
            variable_end_string=')s',
            keep_trailing_newline=True,
            trim_blocks=True,
            lstrip_blocks=True,
        )

    @cached_property
    def with_macos(self):
        return self._set_python_config_value('macos')

    @cached_property
    def with_windows(self):
        return self._set_python_config_value('windows')

    @cached_property
    def with_pypy(self):
        return self._set_python_config_value('pypy')

    @cached_property
    def with_future_python(self):
        value = self._set_python_config_value('future-python')

        if not FUTURE_PYTHON_VERSION:
            return 'false'

        return value

    @cached_property
    def with_docs(self):
        return self._set_python_config_value('docs')

    @cached_property
    def with_sphinx_doctests(self):
        return self._set_python_config_value('sphinx-doctests')

    @cached_property
    def coverage_run_source(self):
        return self.meta_cfg['coverage-run'].get('source', self.path.name)

    @cached_property
    def coverage_fail_under(self):
        return self.meta_cfg['coverage'].setdefault('fail-under', 0)

    @cached_property
    def branch_name(self):
        return get_branch_name(self.args.branch_name, self.config_type)

    def _add_project_to_config_type_list(self):
        """Add the current project to packages.txt if it is not there"""
        if self.override_paths:
            packages_txt_folder = self.override_paths[0]
        else:
            packages_txt_folder = self.config_type_path

        if not (packages_txt_folder / 'packages.txt').exists():
            (packages_txt_folder / 'packages.txt').touch(mode=0o664)

        with open(packages_txt_folder / 'packages.txt') as f:
            known_packages = f.read().splitlines()

        if self.path.name in known_packages:
            print(f'{self.path.name} is already configured '
                  'for this config type, updating.')
        else:
            print(f'{self.path.name} is not yet configured '
                  'for this config type, adding.')
            with open(packages_txt_folder / 'packages.txt', 'a') as f:
                f.write(f'{self.path.name}\n')

    def _set_python_config_value(self, name, default=False):
        """Get value from either python section in config file or cmd line arg.

        Return the value and store it in `meta_cfg`.
        """
        key = f'with-{name}'
        existing_value = self.meta_cfg['python'].get(key, default)
        arg_value = getattr(self.args, key.replace('-', '_'))
        new_value = existing_value or arg_value
        self.meta_cfg['python'][key] = new_value
        return new_value

    def _clean_up_old_settings(self):
        try:
            del self.meta_cfg['python']['with-legacy-python']
        except KeyError:
            pass

    def setup_cfg(self):
        """Copy setup.cfg file to the package being configured."""
        flake8_additional_config = self.cfg_option(
            'flake8', 'additional-config')
        check_manifest_additional_ign = self.cfg_option(
            'check-manifest', 'additional-ignores')
        check_manifest_ignore_bad_ideas = self.cfg_option(
            'check-manifest', 'ignore-bad-ideas')
        isort_known_third_party = prepend_space(
            self.cfg_option('isort', 'known_third_party',
                            default='docutils, pkg_resources, pytz'))
        isort_known_zope = prepend_space(
            self.cfg_option('isort', 'known_zope', default=''))
        isort_known_first_party = prepend_space(
            self.cfg_option('isort', 'known_first_party', default=''))
        isort_known_local_folder = prepend_space(
            self.meta_cfg['isort'].get('known_local_folder', ''))
        isort_additional_config = self.cfg_option(
            'isort', 'additional-config')

        zest_releaser_options = self.meta_cfg['zest-releaser'].get(
            'options', [])
        if self.config_type == 'c-code':
            zest_releaser_options.append('create-wheel = no')

        self.copy_with_meta(
            'setup.cfg.j2',
            self.path / 'setup.cfg',
            self.config_type,
            flake8_additional_config=flake8_additional_config,
            check_manifest_additional_ignores=check_manifest_additional_ign,
            check_manifest_ignore_bad_ideas=check_manifest_ignore_bad_ideas,
            isort_known_third_party=isort_known_third_party,
            isort_known_zope=isort_known_zope,
            isort_known_first_party=isort_known_first_party,
            isort_known_local_folder=isort_known_local_folder,
            isort_additional_config=isort_additional_config,
            with_docs=self.with_docs,
            with_sphinx_doctests=self.with_sphinx_doctests,
            zest_releaser_options=zest_releaser_options,
        )

    def setup_py(self):
        """Update setup.py to current texts."""
        setup_py = self.path / 'setup.py'
        setup_py_content = setup_py.read_text()
        for src, dest in SETUP_PY_REPLACEMENTS.items():
            setup_py_content = setup_py_content.replace(src, dest)
        setup_py.write_text(setup_py_content)

    def gitignore(self):
        git_ignore = self.meta_cfg['git'].get('ignore', [])

        self.copy_with_meta(
            'gitignore.j2', self.path / '.gitignore',
            self.config_type,
            git_ignore=git_ignore,
        )

    def pre_commit_config_yaml(self):
        teyit_exclude = self.meta_cfg["pre-commit"].get("teyit-exclude", "")
        pyupgrade_exclude = self.meta_cfg["pre-commit"].get(
            "pyupgrade-exclude", "")

        self.copy_with_meta(
            "pre-commit-config.yaml.j2",
            self.path / ".pre-commit-config.yaml",
            self.config_type,
            oldest_python_version=self.oldest_python.replace(".", ""),
            teyit_exclude=teyit_exclude,
            pyupgrade_exclude=pyupgrade_exclude,
        )

    def readthedocs(self):
        rtd_build_extra = self.cfg_option(
            'readthedocs', 'build-extra', default=[])
        self.copy_with_meta(
            'readthedocs.yaml.j2',
            self.path / '.readthedocs.yaml',
            self.config_type,
            rtd_build_extra=rtd_build_extra,
        )

    def manylinux_sh(self):
        """Add the scripts to produce binary wheels"""
        manylinux_install_setup = self.meta_cfg['c-code'].get(
            'manylinux-install-setup', [])
        manylinux_aarch64_tests = self.meta_cfg['c-code'].get(
            'manylinux-aarch64-tests', [
                'cd /io/',
                '${PYBIN}/pip install tox',
                'TOXENV=$(tox_env_map "${PYBIN}")',
                '${PYBIN}/tox -e ${TOXENV}',
                'cd ..',
            ])

        if self.template_exists('manylinux.sh'):
            self.copy_with_meta(
                'manylinux.sh', self.path / '.manylinux.sh', self.config_type)
            (self.path / '.manylinux.sh').chmod(0o755)
            stop_at = None
            if not self.with_future_python:
                stop_at = NEWEST_PYTHON_SHORTVERSION
            pkg_name_pattern = re.sub(r"[-_.]+", "?", self.path.name).lower()
            self.copy_with_meta(
                'manylinux-install.sh.j2', self.path / '.manylinux-install.sh',
                self.config_type,
                package_name=self.path.name,
                package_name_pattern=pkg_name_pattern,
                manylinux_install_setup=manylinux_install_setup,
                manylinux_aarch64_tests=manylinux_aarch64_tests,
                with_future_python=self.with_future_python,
                future_python_shortversion=FUTURE_PYTHON_SHORTVERSION,
                supported_python_versions=supported_python_versions(
                    self.oldest_python, short_version=True),
                stop_at=stop_at,
            )
            (self.path / '.manylinux-install.sh').chmod(0o755)
            self.add_manylinux = True

    def cfg_option(self, section, name, default=DEFAULT):
        """Read a value from `self.meta_cfg`, default to `[]` if not existing.
        """
        if default == DEFAULT:
            default = []
        return self.meta_cfg[section].get(name, default)

    def tox_option(self, name, default=DEFAULT):
        """Read a value from the tox options.

        Default not existing ones to `[]`.
        """
        return self.cfg_option('tox', name, default)

    def gh_option(self, name, default=DEFAULT):
        """Read a value from the GitHub actions options.

        Default not existing ones to `[]`.
        """
        return self.cfg_option('github-actions', name, default)

    def tox(self):
        toml_doc = get_pyproject_toml(self.path / 'pyproject.toml')
        build_requirements = toml_doc['build-system'].get('requires', [])
        additional_envlist = self.tox_option('additional-envlist')
        testenv_additional = self.tox_option('testenv-additional')
        testenv_additional_extras = self.tox_option(
            'testenv-additional-extras')
        testenv_commands_pre = self.tox_option('testenv-commands-pre')
        testenv_commands = self.tox_option('testenv-commands')
        testenv_setenv = self.tox_option('testenv-setenv')
        coverage_basepython = self.tox_option(
            'coverage-basepython', default='python3')
        coverage_command = self.tox_option('coverage-command')
        if isinstance(coverage_command, str):
            coverage_command = [coverage_command]
        coverage_additional = self.tox_option('coverage-additional')
        testenv_deps = self.tox_option('testenv-deps')
        coverage_setenv = self.tox_option('coverage-setenv')
        flake8_additional_sources = self.meta_cfg['flake8'].get(
            'additional-sources', '')
        if flake8_additional_sources:
            # Avoid whitespace at end of line
            # if no additional sources are provided:
            flake8_additional_sources = f' {flake8_additional_sources}'
        isort_additional_sources = self.meta_cfg['isort'].get(
            'additional-sources', '')
        if isort_additional_sources:
            # Avoid whitespace at end of line
            # if no additional sources are provided:
            isort_additional_sources = f' {isort_additional_sources}'
        docs_deps = self.tox_option("docs-deps", default=[])
        self.copy_with_meta(
            'tox.ini.j2',
            self.path / 'tox.ini',
            self.config_type,
            additional_envlist=additional_envlist,
            coverage_additional=coverage_additional,
            coverage_basepython=coverage_basepython,
            coverage_command=coverage_command,
            coverage_run_source=self.coverage_run_source,
            coverage_setenv=coverage_setenv,
            coverage_fail_under=self.coverage_fail_under,
            flake8_additional_sources=flake8_additional_sources,
            isort_additional_sources=isort_additional_sources,
            testenv_additional=testenv_additional,
            testenv_additional_extras=testenv_additional_extras,
            testenv_commands=testenv_commands,
            testenv_commands_pre=testenv_commands_pre,
            testenv_deps=testenv_deps,
            testenv_setenv=testenv_setenv,
            with_docs=self.with_docs,
            with_future_python=self.with_future_python,
            with_pypy=self.with_pypy,
            with_sphinx_doctests=self.with_sphinx_doctests,
            docs_deps=docs_deps,
            setuptools_version_spec=SETUPTOOLS_VERSION_SPEC,
            future_python_shortversion=FUTURE_PYTHON_SHORTVERSION,
            supported_python_versions=supported_python_versions(
                self.oldest_python, short_version=True),
            build_requirements=build_requirements,
        )

    def tests_yml(self):
        workflows = self.path / '.github' / 'workflows'
        workflows.mkdir(parents=True, exist_ok=True)

        gha_services = self.gh_option('services')
        gha_additional_config = self.gh_option('additional-config')
        gha_additional_exclude = self.gh_option('additional-exclude')
        gha_steps_before_checkout = self.gh_option('steps-before-checkout')
        gha_additional_install = self.gh_option('additional-install')
        gha_test_environment = self.gh_option('test-environment')
        gha_test_commands = self.gh_option('test-commands')
        py_version_matrix = [
            x for x in zip(supported_python_versions(self.oldest_python,
                                                     short_version=False),
                           supported_python_versions(self.oldest_python,
                                                     short_version=True))]
        self.copy_with_meta(
            'tests.yml.j2',
            workflows / 'tests.yml',
            self.config_type,
            package_name=self.path.name,
            gha_additional_config=gha_additional_config,
            gha_additional_exclude=gha_additional_exclude,
            gha_additional_install=gha_additional_install,
            gha_test_environment=gha_test_environment,
            gha_test_commands=gha_test_commands,
            gha_services=gha_services,
            gha_steps_before_checkout=gha_steps_before_checkout,
            with_docs=self.with_docs,
            with_sphinx_doctests=self.with_sphinx_doctests,
            with_future_python=self.with_future_python,
            future_python_version=FUTURE_PYTHON_VERSION,
            with_pypy=self.with_pypy,
            with_macos=self.with_macos,
            with_windows=self.with_windows,
            manylinux_python_version=MANYLINUX_PYTHON_VERSION,
            manylinux_aarch64=MANYLINUX_AARCH64,
            manylinux_i686=MANYLINUX_I686,
            manylinux_x86_64=MANYLINUX_X86_64,
            pypy_version=PYPY_VERSION,
            setuptools_version_spec=SETUPTOOLS_VERSION_SPEC,
            future_python_shortversion=FUTURE_PYTHON_SHORTVERSION,
            supported_python_versions=py_version_matrix,
        )

    def pre_commit_yml(self):
        workflows = self.path / ".github" / "workflows"
        workflows.mkdir(parents=True, exist_ok=True)

        self.copy_with_meta(
            "pre-commit.yml.j2",
            workflows / "pre-commit.yml",
            self.config_type,
            package_name=self.path.name,
        )

    def manifest_in(self):
        """Modify MANIFEST.in with meta options."""
        manifest_additional_rules = self.meta_cfg['manifest'].get(
            'additional-rules', [])
        if (self.with_docs and 'include *.yaml'
                not in manifest_additional_rules):
            manifest_additional_rules.insert(0, 'include *.yaml')
        if self.config_type == 'c-code' \
                and 'include *.sh' not in manifest_additional_rules:
            manifest_additional_rules.insert(0, 'include *.sh')
        if self.config_type != 'toolkit':
            self.copy_with_meta(
                'MANIFEST.in.j2', self.path / 'MANIFEST.in', self.config_type,
                manifest_additional_rules=manifest_additional_rules,
                with_docs=self.with_docs,
                have_md_files=list(self.path.glob('*.md')),
                have_docs_txt_files=list(self.path.glob('docs/*.txt')),
                have_src_folder=(self.path / 'src').exists())

    def pyproject_toml(self):
        """Modify pyproject.toml with meta options."""
        toml_path = self.path / 'pyproject.toml'
        toml_doc = get_pyproject_toml(
            toml_path,
            comment=META_HINT.format(config_type=self.config_type))

        # Capture some pre-transformation data
        old_requires = toml_doc.get('build-system', {}).get('requires', [])
        old_tools = toml_doc.get('tool', {})

        # Apply template-dependent defaults
        toml_defaults = self.render_with_meta(
            'pyproject_defaults.toml.j2',
            self.config_type,
            setuptools_version_spec=SETUPTOOLS_VERSION_SPEC,
            coverage_run_source=self.coverage_run_source,
            coverage_fail_under=self.coverage_fail_under,
        )
        toml_doc.update(tomlkit.loads(toml_defaults))

        # Create or update section "build-system", we want to control a few
        # build requirement specifications like wheel and setuptools.
        if old_requires:
            setuptools_requirement = [
                x for x in old_requires if x.startswith('setuptools')]
            for setuptools_req in setuptools_requirement:
                old_requires.remove(setuptools_req)
            wheel_requirement = [
                x for x in old_requires if x.startswith('wheel')]
            [old_requires.remove(x) for x in wheel_requirement]
            toml_doc['build-system']['requires'].extend(old_requires)

        # Fix up tool sections because the call to ``update`` above will
        # replace the entire ``tool`` mapping...
        for k, v in old_tools.items():
            if k != 'coverage':  # `coverage` settings are handled by template
                toml_doc['tool'][k] = v

        # Update coverage-related data
        coverage = toml_doc['tool']['coverage']
        add_cfg = self.meta_cfg['coverage-run'].get('additional-config', [])
        for key, value in parse_additional_config(add_cfg).items():
            coverage['run'][key] = value
        omit = self.meta_cfg['coverage-run'].get('omit', [])
        if omit:
            coverage['run']['omit'] = omit

        # Remove empty sections
        for key, value in toml_doc.items():
            if not value:
                toml_doc.remove(key)

        # Fix formatting for some items, especially long lists, so diffs
        # become readable.
        toml_doc['build-system']['requires'].multiline(True)
        toml_doc['tool']['coverage']['report']['exclude_lines'].multiline(True)
        if toml_doc['tool']['coverage'].get('paths'):
            toml_doc['tool']['coverage']['paths']['source'].multiline(True)

        with open(toml_path, 'w') as fp:
            tomlkit.dump(toml_doc, fp, sort_keys=True)

    def render_with_meta(self, template_name, config_type, **kw):
        """Read and render a Jinja template source file"""
        template = self.jinja_env.get_template(template_name)
        return template.render(config_type=config_type, **kw)

    def copy_with_meta(
            self, template_name, destination, config_type,
            meta_hint=META_HINT, **kw):
        """Copy the source file to destination and a hint of origin.

        If kwargs are given they are used as template arguments.

        If the rendered template output is an empty string, don't write it
        to disk. This allows package maintainers to prevent adding certain
        optional files by specifying a custom templates path with the
        ``--template-overrides`` option and adding an empty template there.
        """
        rendered = self.render_with_meta(template_name, config_type, **kw)

        # If the rendered template is empty, give up and return.
        if not rendered.strip():
            return

        meta_hint = meta_hint.format(config_type=config_type)
        if rendered.startswith('#!'):
            she_bang, _, body = rendered.partition('\n')
            content = '\n'.join([she_bang, meta_hint, body])
        else:
            content = '\n'.join([meta_hint, rendered])

        with open(destination, 'w') as f_:
            f_.write(content)

    def configure(self):
        self._add_project_to_config_type_list()
        self._clean_up_old_settings()

        if self.with_sphinx_doctests and not self.with_docs:
            print("The package is configured without sphinx docs, "
                  "but with sphinx doctests.  Is this a mistake?")

        if self.with_docs:
            self.readthedocs()

        self.pyproject_toml()
        self.setup_cfg()
        self.setup_py()
        self.gitignore()
        self.pre_commit_config_yaml()
        self.copy_with_meta(
            'editorconfig.txt', self.path / '.editorconfig', self.config_type)
        self.copy_with_meta(
            'CONTRIBUTING.md', self.path / 'CONTRIBUTING.md', self.config_type,
            meta_hint=META_HINT_MARKDOWN)

        if self.args.commit:
            with change_dir(self.path):
                # We have to add it here otherwise the linter complains
                # that it is not added.
                early_add_candidates = (
                    '.pre-commit-config.yaml',
                    'pyproject.toml',
                    'CONTRIBUTING.md')
                early_add = [x for x in early_add_candidates
                             if pathlib.Path(x).exists()]
                call('git', 'add', *early_add)

        self.manylinux_sh()
        self.tox()
        self.tests_yml()
        self.pre_commit_yml()
        self.manifest_in()

        with change_dir(self.path) as cwd:
            if pathlib.Path('bootstrap.py').exists():
                call('git', 'rm', 'bootstrap.py')
            if pathlib.Path('.travis.yml').exists():
                call('git', 'rm', '.travis.yml')
            if pathlib.Path('.coveragerc').exists():
                call('git', 'rm', '.coveragerc')
            if pathlib.Path('appveyor.yml').exists():
                call('git', 'rm', 'appveyor.yml')
            if self.with_docs and \
               pathlib.Path('.readthedocs.yaml').exists() and \
               self.args.commit:
                call('git', 'add', '.readthedocs.yaml')
            if self.add_manylinux and self.args.commit:
                call('git', 'add', '.manylinux.sh', '.manylinux-install.sh')
            # Remove empty sections:
            meta_cfg = {k: v for k, v in self.meta_cfg.items() if v}
            with open('.meta.toml', 'w') as meta_f:
                meta_f.write(META_HINT.format(config_type=self.config_type))
                meta_f.write('\n')
                tomlkit.dump(meta_cfg, meta_f)

            if self.args.run_tests:
                tox_path = shutil.which('tox') or (
                    pathlib.Path(cwd) / 'bin' / 'tox')
                call(tox_path, '-p', 'auto')

            updating = git_branch(self.branch_name)

            if not self.coverage_fail_under:
                print(
                    'In .meta.toml in section [coverage] the option '
                    '"fail-under" is  0. Please enter a valid minimum '
                    'coverage and rerun.')
                abort(1)
            to_add = [
                ".editorconfig",
                ".github/workflows/tests.yml",
                ".github/workflows/pre-commit.yml",
                ".gitignore",
                ".meta.toml",
                "setup.cfg",
                "tox.ini",
            ]
            if self.config_type != 'toolkit':
                to_add.append('MANIFEST.in')
            if self.args.commit:
                call('git', 'add', *to_add)
                if self.args.commit_msg:
                    commit_msg = self.args.commit_msg
                else:
                    commit_msg = f'Configuring for {self.config_type}'
                call('git', 'commit', '-m', commit_msg)
                if self.args.push:
                    call('git', 'push', '--set-upstream',
                         'origin', self.branch_name)
            print()
            print('If you are an admin and are logged in via `gh auth login`')
            print('update branch protection rules? (y/N)?', end=' ')
            if input().lower() == 'y':
                remote_url = call(
                    'git', 'config', '--get', 'remote.origin.url',
                    capture_output=True).stdout.strip()
                package_name = remote_url.rsplit('/')[-1].removesuffix('.git')
                success = set_branch_protection(
                    package_name, self.path / '.meta.toml')
                if success:
                    print('Successfully updated branch protection rules.')
                else:
                    abort(-1)
            print()
            print('If everything went fine up to here:')
            if updating:
                print('Updated the previously created PR.')
            else:
                print('Create a PR, using the URL shown above.')


def main():
    args = handle_command_line_arguments()

    package = PackageConfiguration(args)
    package.configure()



##############################################################################
#
# Copyright (c) 2022 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import pathlib

from .call import call
from .path import change_dir


def get_commit_id():
    """Return the first 8 digits of the commit id of this repository."""
    with change_dir(pathlib.Path(__file__).parent):
        return call(
            'git', 'rev-parse', '--short=8', 'HEAD',
            capture_output=True).stdout.strip()


def get_branch_name(override, config_type):
    """Get the default branch name but prefer override if not empty.

    The commit ID is based on the meta repository.
    """
    return (
        override
        or f"config-with-{config_type}-template-{get_commit_id()}")


def git_branch(branch_name) -> bool:
    """Switch to existing or create new branch.

    Return `True` if updating.
    """
    branches = call(
        'git', 'branch', '--format', '%(refname:short)',
        capture_output=True).stdout.splitlines()
    if branch_name in branches:
        call('git', 'checkout', branch_name)
        updating = True
    else:
        call('git', 'checkout', '-b', branch_name)
        updating = False
    return updating



#!/usr/bin/env python3
##############################################################################
#
# Copyright (c) 2025 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import argparse
import pathlib
import shutil

from .shared.call import call
from .shared.git import git_branch
from .shared.path import change_dir


def main():
    parser = argparse.ArgumentParser(
        description="Update a repository to PEP 420 native namespace."
    )
    parser.add_argument(
        "path", type=pathlib.Path, help="path to the repository to be updated"
    )
    parser.add_argument(
        "--branch",
        dest="branch_name",
        default=None,
        help="Define a git branch name to be used for the changes. If not"
        " given it is constructed automatically and includes the configuration"
        " type",
    )
    parser.add_argument(
        "--no-commit",
        dest="commit",
        action="store_false",
        default=True,
        help='Don\'t "git commit" changes made by this script.',
    )
    parser.add_argument(
        "--interactive",
        dest="interactive",
        action="store_true",
        default=False,
        help="Run interactively: Scripts will prompt for input. Implies "
        "--no-commit, changes will not be committed and pushed automatically.",
    )
    parser.add_argument(
        '--no-tests',
        dest='run_tests',
        action='store_false',
        default=True,
        help='Skip running unit tests.',
    )

    args = parser.parse_args()
    path = args.path.absolute()

    if not (path / ".git").exists():
        raise ValueError(
            "`path` does not point to a git clone of a repository!")
    if not (path / ".meta.toml").exists():
        raise ValueError("The repository `path` points to has no .meta.toml!")

    with change_dir(path) as cwd_str:
        cwd = pathlib.Path(cwd_str)
        bin_dir = cwd / "bin"
        branch_name = args.branch_name or "pep-420-native-namespace"
        updating = git_branch(branch_name)

        non_interactive_params = []
        if not args.interactive and args.commit:
            non_interactive_params = ["--no-input"]
        else:
            args.commit = False

        call(bin_dir / "bumpversion", "--breaking", *non_interactive_params)
        call(
            bin_dir / "addchangelogentry",
            "Replace ``pkg_resources`` namespace with PEP 420"
            " native namespace."
        )

        setup_py = []
        for line in (path / "setup.py").read_text().splitlines():
            if "from setuptools import find_packages" in line:
                continue
            elif "namespace_packages" in line:
                continue
            elif "packages=" in line:
                continue
            elif "package_dir=" in line:
                continue
            elif "zope.testrunner" in line:
                setup_py.append(
                    line.replace("zope.testrunner", "zope.testrunner >= 6.4")
                )
            else:
                setup_py.append(line)

        (path / "setup.py").write_text("\n".join(setup_py) + "\n")

        for src_dir_cont in (path / "src").iterdir():
            if not src_dir_cont.is_dir():
                continue
            pkg_init = src_dir_cont / "__init__.py"
            if pkg_init.exists():
                pkg_init.unlink()
            for pkg_dir_cont in src_dir_cont.iterdir():
                if not pkg_dir_cont.is_dir():
                    continue
                sub_pkg_init = pkg_dir_cont / "__init__.py"
                if sub_pkg_init.exists():
                    if "pkg_resources" in sub_pkg_init.read_text():
                        sub_pkg_init.unlink()

        if args.commit:
            print("Adding all changes ...")
            call("git", "add", ".")

        if args.run_tests:
            tox_path = shutil.which("tox") or (cwd / "bin" / "tox")
            call(tox_path, "-p", "auto")

        if args.commit:
            print("Committing and pushing all changes ...")
            call("git", "commit", "-m", "Switch to PEP 420 native namespace.")
            call("git", "push", "--set-upstream", "origin", branch_name)
            if updating:
                print("Updated the previously created PR.")
            else:
                print(
                    "Are you logged in via `gh auth login` to create a PR?"
                    " (y/N)?", end=" ",
                )
                if input().lower() == "y":
                    call(
                        "gh",
                        "pr",
                        "create",
                        "--fill",
                        "--title",
                        "Switch to PEP 420 native namespace.",
                    )
                else:
                    print("If everything went fine up to here:")
                    print("Create a PR, using the URL shown above.")
        else:
            print("Applied all changes. Please check and commit manually.")



#!/usr/bin/env python3
##############################################################################
#
# Copyright (c) 2020 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import argparse
import sys

from .shared.call import call
from .shared.packages import list_packages
from .shared.path import change_dir
from .shared.path import path_factory


def main():
    parser = argparse.ArgumentParser(
        description='Call a script on all repositories listed'
                    ' in a packages.txt.',
        epilog='Additional optional arguments are passed'
               ' directly to the script.')
    parser.add_argument(
        'script', type=path_factory('script', has_extension='.py'),
        help='path to the Python script to be called')
    parser.add_argument(
        'packages_txt',
        type=path_factory(
            'packages.txt',
            has_extension='.txt'),
        help='path to packages.txt; script is called on each repository listed'
        ' inside',
        metavar='packages.txt')
    parser.add_argument(
        'clones', type=path_factory('clones', is_dir=True),
        help='path to the directory where the clones of the repositories are'
             ' stored')

    # idea from https://stackoverflow.com/a/37367814/8531312
    args, sub_args = parser.parse_known_args()
    packages = list_packages(args.packages_txt)

    for package in packages:
        print(f'*** Running {args.script.name} on {package} ***')
        if (args.clones / package).exists():
            with change_dir(args.clones / package):
                print('Updating existing checkout …')
                call('git', 'stash')
                call('git', 'checkout', 'master')
                call('git', 'pull')
        else:
            with change_dir(args.clones):
                print('Cloning repository …')
                call('git', 'clone',
                     f'https://github.com/zopefoundation/{package}')

        call_args = [
            sys.executable,
            args.script,
            args.clones / package
        ]
        call_args.extend(sub_args)
        call(*call_args)



##############################################################################
#
# Copyright (c) 2021 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import configparser
import itertools
import pathlib

import tomlkit
from packaging.version import parse as parse_version
from tomlkit.toml_document import TOMLDocument


TYPES = ['buildout-recipe', 'c-code', 'pure-python', 'zope-product', 'toolkit']
ORG = 'zopefoundation'
BASE_PATH = pathlib.Path(__file__).parent.parent
OLDEST_PYTHON_VERSION = '3.10'
NEWEST_PYTHON_VERSION = '3.14'
FUTURE_PYTHON_VERSION = '3.15'
PYPY_VERSION = '3.11'  # pypy-3.11
SETUPTOOLS_VERSION_SPEC = '>= 78.1.1,< 81'
MANYLINUX_PYTHON_VERSION = '3.11'
MANYLINUX_AARCH64 = 'manylinux2014_aarch64'
MANYLINUX_I686 = 'manylinux2014_i686'
MANYLINUX_X86_64 = 'manylinux2014_x86_64'


def get_pyproject_toml(path: pathlib.Path, comment='') -> TOMLDocument:
    """Parse ``pyproject.toml`` and return its values as ``TOMLDocument``.

    Args:
        path (str, pathlib.Path): Filesystem path to a pyproject.toml file.

    Kwargs:
        comment (str): Optional comment added to the top of the file.

    Returns:
        A TOMLDocument instance from the pyproject.toml file.
    """
    if path.exists():
        with open(path) as fp:
            toml_contents = fp.read()
    else:
        toml_contents = ''

    if comment and not\
       (toml_contents.startswith(comment) or
            toml_contents.startswith(f'# \n{comment}')):
        toml_contents = f'{comment}\n{toml_contents}'

    return tomlkit.loads(toml_contents)


def parse_additional_config(cfg):
    """Attempt to parse "additional-config" data

    "additional-config" sections usually contain ini-style key/value pairs
    packaged into a sequence of lines.

    Returns a mapping of keys and values found
    """
    data = {}

    if cfg:
        parser = configparser.ConfigParser()
        parser.read_string('[dummysection]\n' + '\n'.join(cfg))
        for key, value in parser.items('dummysection'):
            if '\n' in value:
                value = value.split()
            else:
                for func in (parser.getboolean,
                             parser.getint,
                             parser.getfloat):
                    try:
                        value = func('dummysection', key)
                        break
                    except Exception:
                        pass
            data[key] = value

    return data


def supported_python_versions(oldest_version=OLDEST_PYTHON_VERSION,
                              short_version=False):
    """Create a list containing all supported Python versions

    Uses the configured oldest and newest Python versions to compute a list
    containing all versions from oldest to newest that can be iterated over in
    the templates.

    Kwargs:
        oldest_version (str):
            The oldest supported Python version, e.g. '3.12'.

        short_version (bool):
            Return short versions like "313" instead of "3.13". Default False.
    """
    minor_versions = []
    oldest_python = parse_version(oldest_version)
    newest_python = parse_version(NEWEST_PYTHON_VERSION)
    for minor in range(oldest_python.minor, newest_python.minor + 1):
        minor_versions.append(minor)
    supported = [f'{oldest_python.major}.{minor}' for minor in minor_versions]

    if short_version:
        return [x.replace('.', '') for x in supported]

    return supported


def list_packages(path: pathlib.Path) -> list:
    """List the packages in ``path``.

    ``path`` must point to a packages.txt file.
    """
    return [
        p for p in path.read_text().split('\n') if p and not p.startswith('#')
    ]


ALL_REPOS = itertools.chain(
    *[list_packages(BASE_PATH / type / 'packages.txt') for type in TYPES])



##############################################################################
#
# Copyright (c) 2020 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import argparse
import contextlib
import os
import pathlib


@contextlib.contextmanager
def change_dir(path):
    """Change current working directory temporarily to `path`.

    Yields the current working directory before the change.
    """
    cwd = os.getcwd()
    try:
        os.chdir(path)
        yield cwd
    finally:
        os.chdir(cwd)


def path_factory(parameter_name, *, has_extension=None, is_dir=False):
    """Return factory creating pathlib.Path object if requirements are matched.

    The factory raises an exception otherwise.
    """
    def factory(str):
        path = pathlib.Path(str)
        if not path.exists():
            raise argparse.ArgumentTypeError(f'{str!r} does not exist!')
        if has_extension is not None and path.suffix != has_extension:
            raise argparse.ArgumentTypeError(
                f'The required extension is {has_extension!r}'
                f' not {path.suffix!r}.')
        if is_dir and not path.is_dir():
            raise argparse.ArgumentTypeError('has to point to a directory!')
        return path
    return factory



##############################################################################
#
# Copyright (c) 2020 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import subprocess
import sys
import textwrap


def abort(exitcode):
    """Ask the user to abort."""
    print('ABORTING: Please fix the errors shown above.')
    print('Proceed anyway (y/N)?', end=' ')
    if input().lower() != 'y':
        sys.exit(exitcode)


def wait_for_accept():
    """Wait until the user has hit enter."""
    print('Proceed by hitting <ENTER>')
    input()


def call(*args, capture_output=False, cwd=None, allowed_return_codes=(0, )):
    """Call `args` as a subprocess.

    If it fails exit the process.
    """
    result = subprocess.run(
        args, capture_output=capture_output, text=True, cwd=cwd)
    if result.returncode not in allowed_return_codes:
        if capture_output:
            abort_text = textwrap.dedent(f'''
                error code: {result.returncode}
                stderr: {result.stderr}
                stdout: {result.stdout}''')
        else:
            abort_text = result.returncode
        abort(abort_text)
    return result



#!/usr/bin/env python3
import argparse
import json
import os
import pathlib
import tempfile

import requests
import tomlkit

from .shared.call import abort
from .shared.call import call
from .shared.packages import ALL_REPOS
from .shared.packages import MANYLINUX_AARCH64
from .shared.packages import MANYLINUX_I686
from .shared.packages import MANYLINUX_PYTHON_VERSION
from .shared.packages import MANYLINUX_X86_64
from .shared.packages import NEWEST_PYTHON_VERSION
from .shared.packages import OLDEST_PYTHON_VERSION
from .shared.packages import ORG
from .shared.packages import PYPY_VERSION


BASE_URL = f'https://raw.githubusercontent.com/{ORG}'
NEWEST_PYTHON = f'py{NEWEST_PYTHON_VERSION.replace(".", "")}'
DEFAULT_BRANCH = 'master'


def _call_gh(
        method, path, repo, *args, capture_output=True,
        allowed_return_codes=(0, )):
    """Call the gh api command."""
    return call(
        'gh', 'api',
        '--method', method,
        '-H', 'Accept: application/vnd.github+json',
        '-H', 'X-GitHub-Api-Version: 2022-11-28',
        f'/repos/{ORG}/{repo}/branches/{DEFAULT_BRANCH}/{path}',
        *args, capture_output=capture_output,
        allowed_return_codes=allowed_return_codes)


def set_branch_protection(
        repo: str, meta_path: pathlib.Path | None = None) -> bool:
    result = _call_gh(
        'GET', 'protection/required_pull_request_reviews', repo,
        allowed_return_codes=(0, 1))
    required_pull_request_reviews = None
    if result.returncode == 1:
        if json.loads(result.stdout)['message'] != "Branch not protected":
            # If there is no branch protection we create it later on using the
            # PUT call, but if there is another error we show it:
            print(result.stdout)
            abort(result.returncode)
    else:
        required_approving_review_count = json.loads(
            result.stdout)['required_approving_review_count']
        required_pull_request_reviews = {
            'required_approving_review_count': required_approving_review_count
        }

    if meta_path is None:
        response = requests.get(
            f'{BASE_URL}/{repo}/{DEFAULT_BRANCH}/.meta.toml', timeout=30)
        meta_toml = tomlkit.loads(response.text)
    else:
        with open(meta_path) as f:
            meta_toml = tomlkit.load(f)
    template = meta_toml['meta']['template']
    with_docs = meta_toml['python'].get('with-docs', False)
    with_pypy = meta_toml['python']['with-pypy']
    oldest_python_version = meta_toml['python'].get('oldest-python',
                                                    OLDEST_PYTHON_VERSION)
    oldest_python = f"py{oldest_python_version.replace('.', '')}"
    with_windows = meta_toml['python']['with-windows']
    with_macos = meta_toml['python']['with-macos']
    required = ['linting']
    if template == 'c-code':
        required.extend([
            f'manylinux ({MANYLINUX_PYTHON_VERSION}, {MANYLINUX_AARCH64})',
            f'manylinux ({MANYLINUX_PYTHON_VERSION}, {MANYLINUX_I686})',
            f'manylinux ({MANYLINUX_PYTHON_VERSION}, {MANYLINUX_X86_64})',
            f'test ({oldest_python_version}, macos-latest)',
            f'test ({NEWEST_PYTHON_VERSION}, macos-latest)',
            f'test ({oldest_python_version}, ubuntu-latest)',
            f'test ({NEWEST_PYTHON_VERSION}, ubuntu-latest)',
            'coveralls_finish',
        ])
        if with_docs:
            required.append(
                f'docs ({MANYLINUX_PYTHON_VERSION}, ubuntu-latest)')
        if with_pypy:
            required.append(f'test (pypy-{PYPY_VERSION}, ubuntu-latest)')
            required.append(f'test (pypy-{PYPY_VERSION}, windows-latest)')
        if with_windows:
            required.extend([
                f'test ({oldest_python_version}, windows-latest)',
                f'test ({NEWEST_PYTHON_VERSION}, windows-latest)',
            ])
    elif with_windows or with_macos:
        required.extend([
            'ubuntu-coverage',
            f'ubuntu-{oldest_python}',
            f'ubuntu-{NEWEST_PYTHON}',
        ])
        if with_windows:
            required.extend([
                f'windows-{oldest_python}',
                f'windows-{NEWEST_PYTHON}',
            ])
        if with_macos:
            required.extend([
                f'macos-{oldest_python}',
                f'macos-{NEWEST_PYTHON}',
            ])
        if with_pypy:
            required.extend([
                'ubuntu-pypy3',
                'windows-pypy3',
            ])
        if with_docs:
            required.append('ubuntu-docs')
    else:  # default for most packages
        required.extend([oldest_python, NEWEST_PYTHON])
        if template != 'toolkit':
            required.append('coverage')
        if with_docs:
            required.append('docs')
        if with_pypy:
            required.append('pypy3')

    data = {
        'allow_deletions': False,
        'allow_force_pushes': False,
        'allow_fork_syncing': True,
        'lock_branch': False,
        'enforce_admins': None,
        'restrictions': None,
        'required_conversation_resolution': True,
        'required_linear_history': False,
        'required_pull_request_reviews': required_pull_request_reviews,
        'required_status_checks': {'contexts': required, 'strict': False}
    }
    fd, filename = tempfile.mkstemp('config.json', 'meta', text=True)
    try:
        file = os.fdopen(fd, 'w')
        json.dump(data, file)
        file.close()
        _call_gh('PUT', 'protection', repo, '--input', filename)
    finally:
        os.unlink(filename)
    return True


def main():
    parser = argparse.ArgumentParser(
        description='Set the branch protection rules for all known packages.\n'
                    'Prerequsites: `gh auth login`.')
    parser.add_argument(
        '--I-am-authenticated',
        help='If you are authenticated via `gh auth login`, use this required'
        ' parameter.',
        action='store_true',
        required=True)
    parser.add_argument(
        '-r', '--repos',
        help='Run the script only for the given repos instead of all.',
        metavar='NAME', nargs='*', default=[])
    parser.add_argument(
        '-m', '--meta',
        help='Use this .meta.toml instead the one on `master` of the repos.',
        metavar='PATH', default=None, type=pathlib.Path)

    args = parser.parse_args()
    repos = args.repos if args.repos else ALL_REPOS
    meta_path = args.meta

    if meta_path and len(repos) > 1:
        print('--meta can only be used together with a single repos.')
        abort(-1)

    for repo in repos:
        print(repo, end="")
        set_branch_protection(repo, meta_path)
        print(' ✅')



##############################################################################
#
# Copyright (c) 2024 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Setup for zope.meta package
"""

import os

from setuptools import setup


def read(*rnames):
    with open(os.path.join(os.path.dirname(__file__), *rnames)) as stream:
        return stream.read()


setup(
    name='zope.meta',
    version='2.0',
    author='Zope Foundation and Contributors',
    author_email='zope-dev@zope.dev',
    description='Helper functions for package management',
    long_description=(
        read('README.rst')
        + '\n\n' +
        read('CHANGES.rst')
    ),
    long_description_content_type='text/x-rst',
    keywords="zope packaging",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Programming Language :: Python :: Implementation :: CPython',
        'Natural Language :: English',
        'Operating System :: OS Independent',
    ],
    license='ZPL-2.1',
    url='https://github.com/zopefoundation/meta',
    project_urls={
        'Documentation': 'https://zopemeta.readthedocs.io',
        'Issue Tracker': 'https://github.com/zopefoundation/meta/issues',
        'Sources': 'https://github.com/zopefoundation/meta',
    },
    install_requires=[
        'setuptools',
        'check-python-versions',
        'Jinja2',
        'packaging',
        'pyupgrade',
        'requests',
        'tomlkit',
        'tox',
        'zest.releaser',
    ],
    python_requires='>=3.10',
    include_package_data=True,
    zip_safe=False,
    extras_require={
        'test': ['zope.testrunner >= 6.4'],
        'docs': ['Sphinx', 'furo'],
    },
    entry_points={
        'console_scripts': [
            'config-package=zope.meta.config_package:main',
            'multi-call=zope.meta.multi_call:main',
            're-enable-actions=zope.meta.re_enable_actions:main',
            (
                'set-branch-protection-rules='
                'zope.meta.set_branch_protection_rules:main'
            ),
            'update-python-support=zope.meta.update_python_support:main',
            'switch-to-pep420=zope.meta.pep_420:main',
        ],
    },
)
